# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from collections.abc import Iterator
from contextlib import contextmanager
from functools import partial
from tempfile import NamedTemporaryFile
from typing import Any

import click
import coverage
import lsst.pipe.base.cli.opt as pipeBaseOpts
from lsst.ctrl.mpexec.showInfo import ShowInfo
from lsst.daf.butler.cli.opt import (
    config_file_option,
    config_option,
    confirm_option,
    options_file_option,
    processes_option,
    repo_argument,
)
from lsst.daf.butler.cli.utils import MWCtxObj, catch_and_exit, option_section, unwrap

from .. import opt as ctrlMpExecOpts
from .. import script
from ..script import confirmable
from ..utils import PipetaskCommand, makePipelineActions

epilog = unwrap(
    """Notes:

--task, --delete, --config, --config-file, and --instrument action options can
appear multiple times; all values are used, in order left to right.

FILE reads command-line options from the specified file. Data may be
distributed among multiple lines (e.g. one option per line). Data after # is
treated as a comment and ignored. Blank lines and lines starting with # are
ignored.)
"""
)


def _collectActions(ctx: click.Context, **kwargs: Any) -> dict[str, Any]:
    """Extract pipeline building options, replace them with PipelineActions,
    return updated `kwargs`.

    Notes
    -----
    The pipeline actions (task, delete, config, config_file, and instrument)
    must be handled in the order they appear on the command line, but the CLI
    specification gives them all different option names. So, instead of using
    the individual action options as they appear in kwargs (because
    invocation order can't be known), we capture the CLI arguments by
    overriding `click.Command.parse_args` and save them in the Context's
    `obj` parameter. We use `makePipelineActions` to create a list of
    pipeline actions from the CLI arguments and pass that list to the script
    function using the `pipeline_actions` kwarg name, and remove the action
    options from kwargs.
    """
    for pipelineAction in (
        ctrlMpExecOpts.task_option.name(),
        ctrlMpExecOpts.delete_option.name(),
        config_option.name(),
        config_file_option.name(),
        pipeBaseOpts.instrument_option.name(),
    ):
        kwargs.pop(pipelineAction)

    actions = makePipelineActions(MWCtxObj.getFrom(ctx).args)
    pipeline_actions = []
    for action in actions:
        pipeline_actions.append(action)

    kwargs["pipeline_actions"] = pipeline_actions
    return kwargs


def _unhandledShow(show: ShowInfo, cmd: str) -> None:
    if show.unhandled:
        print(
            f"The following '--show' options were not known to the {cmd} command: "
            f"{', '.join(show.unhandled)}",
            file=sys.stderr,
        )


@click.command(cls=PipetaskCommand, epilog=epilog, short_help="Build pipeline definition.")
@click.pass_context
@ctrlMpExecOpts.show_option()
@ctrlMpExecOpts.pipeline_build_options()
@option_section(sectionText="")
@options_file_option()
@catch_and_exit
def build(ctx: click.Context, **kwargs: Any) -> None:
    """Build and optionally save pipeline definition.

    This does not require input data to be specified.
    """
    kwargs = _collectActions(ctx, **kwargs)
    show = ShowInfo(kwargs.pop("show", []))
    if kwargs.get("butler_config") is not None and {"pipeline-graph", "task-graph"}.isdisjoint(show.commands):
        raise click.ClickException(
            "--butler-config was provided but nothing uses it "
            "(only --show pipeline-graph and --show task-graph do)."
        )
    script.build(**kwargs, show=show)
    _unhandledShow(show, "build")


@contextmanager
def coverage_context(kwargs: dict[str, Any]) -> Iterator[None]:
    """Enable coverage recording."""
    packages = kwargs.pop("cov_packages", ())
    report = kwargs.pop("cov_report", True)
    if not kwargs.pop("coverage", False):
        yield
        return
    with NamedTemporaryFile("w") as rcfile:
        rcfile.write(
            """
[run]
branch = True
concurrency = multiprocessing
"""
        )
        if packages:
            packages_str = ",".join(packages)
            rcfile.write(f"source_pkgs = {packages_str}\n")
        rcfile.flush()
        cov = coverage.Coverage(config_file=rcfile.name)
        cov.start()
        try:
            yield
        finally:
            cov.stop()
            cov.save()
            if report:
                outdir = "./covhtml"
                cov.html_report(directory=outdir)
                click.echo(f"Coverage report written to {outdir}.")


@click.command(cls=PipetaskCommand, epilog=epilog)
@click.pass_context
@ctrlMpExecOpts.show_option()
@ctrlMpExecOpts.pipeline_build_options()
@ctrlMpExecOpts.qgraph_options()
@ctrlMpExecOpts.butler_options()
@option_section(sectionText="")
@options_file_option()
@catch_and_exit
def qgraph(ctx: click.Context, **kwargs: Any) -> None:
    """Build and optionally save quantum graph."""
    kwargs = _collectActions(ctx, **kwargs)
    with coverage_context(kwargs):
        show = ShowInfo(kwargs.pop("show", []))
        pipeline = script.build(**kwargs, show=show)
        if show.handled and not show.unhandled:
            print(
                "No quantum graph generated. The --show option was given and all options were processed.",
                file=sys.stderr,
            )
            return
        if script.qgraph(pipelineObj=pipeline, **kwargs, show=show) is None:
            raise click.ClickException("QuantumGraph was empty; CRITICAL logs above should provide details.")
        _unhandledShow(show, "qgraph")


@click.command(cls=PipetaskCommand, epilog=epilog)
@ctrlMpExecOpts.run_options()
@catch_and_exit
def run(ctx: click.Context, **kwargs: Any) -> None:
    """Build and execute pipeline and quantum graph."""
    kwargs = _collectActions(ctx, **kwargs)
    with coverage_context(kwargs):
        show = ShowInfo(kwargs.pop("show", []))
        pipeline = script.build(**kwargs, show=show)
        if show.handled and not show.unhandled:
            print(
                "No quantum graph generated or pipeline executed. "
                "The --show option was given and all options were processed.",
                file=sys.stderr,
            )
            return
        if (qgraph := script.qgraph(pipelineObj=pipeline, **kwargs, show=show)) is None:
            raise click.ClickException("QuantumGraph was empty; CRITICAL logs above should provide details.")
        _unhandledShow(show, "run")
        if show.handled:
            print(
                "No pipeline executed. The --show option was given and all options were processed.",
                file=sys.stderr,
            )
            return
        script.run(qgraphObj=qgraph, **kwargs)


@click.command(cls=PipetaskCommand)
@ctrlMpExecOpts.butler_config_option()
@ctrlMpExecOpts.collection_argument()
@confirm_option()
@ctrlMpExecOpts.recursive_option(
    help="""If the parent CHAINED collection has child CHAINED collections,
    search the children until nested chains that start with the parent's name
    are removed."""
)
def purge(confirm: bool, **kwargs: Any) -> None:
    """Remove a CHAINED collection and its contained collections.

    COLLECTION is the name of the chained collection to purge. it must not be a
    child of any other CHAINED collections

    Child collections must be members of exactly one collection.

    The collections that will be removed will be printed, there will be an
    option to continue or abort (unless using --no-confirm).
    """
    confirmable.confirm(partial(script.purge, **kwargs), confirm)


@click.command(cls=PipetaskCommand)
@ctrlMpExecOpts.butler_config_option()
@ctrlMpExecOpts.collection_argument()
@confirm_option()
def cleanup(confirm: bool, **kwargs: Any) -> None:
    """Remove non-members of CHAINED collections.

    Removes collections that start with the same name as a CHAINED
    collection but are not members of that collection.
    """
    confirmable.confirm(partial(script.cleanup, **kwargs), confirm)


@click.command(cls=PipetaskCommand)
@repo_argument()
@ctrlMpExecOpts.qgraph_argument()
@ctrlMpExecOpts.config_search_path_option()
@ctrlMpExecOpts.qgraph_id_option()
@ctrlMpExecOpts.coverage_options()
def pre_exec_init_qbb(repo: str, qgraph: str, **kwargs: Any) -> None:
    """Execute pre-exec-init on Quantum-Backed Butler.

    REPO is the location of the butler/registry config file.

    QGRAPH is the path to a serialized Quantum Graph file.
    """
    with coverage_context(kwargs):
        script.pre_exec_init_qbb(repo, qgraph, **kwargs)


@click.command(cls=PipetaskCommand)
@repo_argument()
@ctrlMpExecOpts.qgraph_argument()
@ctrlMpExecOpts.config_search_path_option()
@ctrlMpExecOpts.qgraph_id_option()
@ctrlMpExecOpts.qgraph_node_id_option()
@processes_option()
@ctrlMpExecOpts.pdb_option()
@ctrlMpExecOpts.profile_option()
@ctrlMpExecOpts.coverage_options()
@ctrlMpExecOpts.debug_option()
@ctrlMpExecOpts.start_method_option()
@ctrlMpExecOpts.timeout_option()
@ctrlMpExecOpts.fail_fast_option()
@ctrlMpExecOpts.summary_option()
@ctrlMpExecOpts.enable_implicit_threading_option()
@ctrlMpExecOpts.cores_per_quantum_option()
@ctrlMpExecOpts.memory_per_quantum_option()
def run_qbb(repo: str, qgraph: str, **kwargs: Any) -> None:
    """Execute pipeline using Quantum-Backed Butler.

    REPO is the location of the butler/registry config file.

    QGRAPH is the path to a serialized Quantum Graph file.
    """
    with coverage_context(kwargs):
        script.run_qbb(repo, qgraph, **kwargs)


@click.command(cls=PipetaskCommand)
@ctrlMpExecOpts.qgraph_argument()
@ctrlMpExecOpts.run_argument()
@ctrlMpExecOpts.output_qgraph_argument()
@ctrlMpExecOpts.metadata_run_key_option()
@ctrlMpExecOpts.update_graph_id_option()
def update_graph_run(
    qgraph: str,
    run: str,
    output_qgraph: str,
    metadata_run_key: str,
    update_graph_id: bool,
) -> None:
    """Update existing quantum graph with new output run name and re-generate
    output dataset IDs.

    QGRAPH is the URL to a serialized Quantum Graph file.

    RUN is the new RUN collection name for output graph.

    OUTPUT_QGRAPH is the URL to store the updated Quantum Graph.
    """
    script.update_graph_run(qgraph, run, output_qgraph, metadata_run_key, update_graph_id)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["TaskFactory"]

import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from lsst.pipe.base import TaskFactory as BaseTaskFactory

if TYPE_CHECKING:
    from lsst.daf.butler import DatasetRef, LimitedButler
    from lsst.pipe.base import PipelineTask, TaskDef

_LOG = logging.getLogger(__name__)


class TaskFactory(BaseTaskFactory):
    """Class instantiating PipelineTasks."""

    def makeTask(
        self, taskDef: TaskDef, butler: LimitedButler, initInputRefs: Iterable[DatasetRef] | None
    ) -> PipelineTask:
        # docstring inherited

        config = taskDef.config

        # Get init inputs from butler.
        init_inputs: dict[str, Any] = {}
        if initInputRefs:
            connections = config.connections.ConnectionsClass(config=config)
            for name in connections.initInputs:
                attribute = getattr(connections, name)
                dataset_type_name = attribute.name
                for ref in initInputRefs:
                    if ref.datasetType.name == dataset_type_name:
                        init_inputs[name] = butler.get(ref)
                        break

        # make task instance
        task = taskDef.taskClass(config=config, initInputs=init_inputs, name=taskDef.label)
        return task



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from types import SimpleNamespace

from ... import CmdLineFwk, TaskFactory


def run_qbb(
    butler_config: str,
    qgraph: str,
    config_search_path: list[str] | None,
    qgraph_id: str | None,
    qgraph_node_id: list[int] | None,
    processes: int,
    pdb: str | None,
    profile: str | None,
    debug: bool,
    start_method: str | None,
    timeout: int | None,
    fail_fast: bool,
    summary: str | None,
    enable_implicit_threading: bool,
    cores_per_quantum: int,
    memory_per_quantum: str,
) -> None:
    """Implement the command line interface ``pipetask run-qbb`` subcommand.

    Should only be called by command line tools and unit test code that tests
    this function.

    Parameters
    ----------
    butler_config : `str`
        The path location of the gen3 butler/registry config file.
    qgraph : `str`
        URI location for a serialized quantum graph definition.
    config_search_path : `list` [`str`]
        Additional search paths for butler configuration.
    qgraph_id : `str` or `None`
        Quantum graph identifier, if specified must match the identifier of the
        graph loaded from a file. Ignored if graph is not loaded from a file.
    qgraph_node_id : `iterable` of `int`, or `None`
        Only load a specified set of nodes if graph is loaded from a file,
        nodes are identified by integer IDs.
    processes : `int`
        The number of processes to use.
    pdb : `str` or `None`
        Debugger to launch for exceptions.
    profile : `str`
        File name to dump cProfile information to.
    debug : `bool`
        If true, enable debugging output using lsstDebug facility (imports
        debug.py).
    start_method : `str` or `None`
        Start method from `multiprocessing` module, `None` selects the best
        one for current platform.
    timeout : `int`
        Timeout for multiprocessing; maximum wall time (sec).
    fail_fast : `bool`
        If true then stop processing at first error, otherwise process as many
        tasks as possible.
    summary : `str` or `None`
        File path to store job report in JSON format.
    enable_implicit_threading : `bool`
        If `True`, do not disable implicit threading by third-party libraries.
        Implicit threading is always disabled during actual quantum execution
        if ``processes > 1``.
    cores_per_quantum : `int`
        Number of cores that can be used by each quantum.
    memory_per_quantum : `str`
        Amount of memory that each quantum can be allowed to use. Empty string
        implies no limit. The string can be either a single integer (implying
        units of MB) or a combination of number and unit.
    """
    args = SimpleNamespace(
        butler_config=butler_config,
        qgraph=qgraph,
        config_search_path=config_search_path,
        qgraph_id=qgraph_id,
        qgraph_node_id=qgraph_node_id,
        processes=processes,
        pdb=pdb,
        profile=profile,
        enableLsstDebug=debug,
        start_method=start_method,
        timeout=timeout,
        fail_fast=fail_fast,
        summary=summary,
        enable_implicit_threading=enable_implicit_threading,
        cores_per_quantum=cores_per_quantum,
        memory_per_quantum=memory_per_quantum,
    )

    f = CmdLineFwk()
    task_factory = TaskFactory()
    f.runGraphQBB(task_factory, args)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import tempfile
import unittest

import lsst.daf.butler
import lsst.daf.butler.tests as butlerTests
import lsst.pex.config
import lsst.utils.tests
from lsst.ctrl.mpexec import SeparablePipelineExecutor
from lsst.pipe.base import Instrument, Pipeline, PipelineDatasetTypes, TaskMetadata
from lsst.resources import ResourcePath

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class SeparablePipelineExecutorTests(lsst.utils.tests.TestCase):
    """Test the SeparablePipelineExecutor API with a trivial task."""

    pipeline_file = os.path.join(TESTDIR, "pipeline_separable.yaml")

    def setUp(self):
        repodir = tempfile.TemporaryDirectory()
        # TemporaryDirectory warns on leaks; addCleanup also keeps it from
        # getting garbage-collected.
        self.addCleanup(tempfile.TemporaryDirectory.cleanup, repodir)

        # standalone parameter forces the returned config to also include
        # the information from the search paths.
        config = lsst.daf.butler.Butler.makeRepo(
            repodir.name, standalone=True, searchPaths=[os.path.join(TESTDIR, "config")]
        )
        butler = lsst.daf.butler.Butler(config, writeable=True)
        output = "fake"
        output_run = f"{output}/{Instrument.makeCollectionTimestamp()}"
        butler.registry.registerCollection(output_run, lsst.daf.butler.CollectionType.RUN)
        butler.registry.registerCollection(output, lsst.daf.butler.CollectionType.CHAINED)
        butler.registry.setCollectionChain(output, [output_run])
        self.butler = lsst.daf.butler.Butler(butler=butler, collections=[output], run=output_run)

        butlerTests.addDatasetType(self.butler, "input", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "intermediate", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "a_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "a_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, "a_config", set(), "Config")

    def test_pre_execute_qgraph(self):
        # Too hard to make a quantum graph from scratch.
        executor = SeparablePipelineExecutor(self.butler)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)

        butlerTests.addDatasetType(self.butler, "output", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "b_config", set(), "Config")
        butlerTests.addDatasetType(self.butler, "b_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "b_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, PipelineDatasetTypes.packagesDatasetName, set(), "Packages")

        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=False,
            save_init_outputs=False,
            save_versions=False,
        )
        self.assertFalse(self.butler.exists("a_config", {}, collections=[self.butler.run]))
        self.assertFalse(self.butler.exists(PipelineDatasetTypes.packagesDatasetName, {}))

    def test_pre_execute_qgraph_unconnected(self):
        # Unconnected graph; see
        # test_make_quantum_graph_nowhere_skippartial_clobber.
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        graph = executor.make_quantum_graph(pipeline)

        butlerTests.addDatasetType(self.butler, "output", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "b_config", set(), "Config")
        butlerTests.addDatasetType(self.butler, "b_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "b_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, PipelineDatasetTypes.packagesDatasetName, set(), "Packages")

        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=False,
            save_init_outputs=False,
            save_versions=False,
        )
        self.assertFalse(self.butler.exists("a_config", {}, collections=[self.butler.run]))
        self.assertFalse(self.butler.exists(PipelineDatasetTypes.packagesDatasetName, {}))

    def test_pre_execute_qgraph_empty(self):
        executor = SeparablePipelineExecutor(self.butler)
        graph = lsst.pipe.base.QuantumGraph({}, universe=self.butler.dimensions)

        butlerTests.addDatasetType(self.butler, "output", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "b_config", set(), "Config")
        butlerTests.addDatasetType(self.butler, "b_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "b_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, PipelineDatasetTypes.packagesDatasetName, set(), "Packages")

        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=False,
            save_init_outputs=False,
            save_versions=False,
        )
        self.assertFalse(self.butler.exists("a_config", {}, collections=[self.butler.run]))
        self.assertFalse(self.butler.exists(PipelineDatasetTypes.packagesDatasetName, {}))

    def test_pre_execute_qgraph_register(self):
        executor = SeparablePipelineExecutor(self.butler)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)

        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=False,
            save_versions=False,
        )
        self.assertEqual({d.name for d in self.butler.registry.queryDatasetTypes("output")}, {"output"})
        self.assertEqual(
            {d.name for d in self.butler.registry.queryDatasetTypes("b_*")},
            {"b_config", "b_log", "b_metadata"},
        )
        self.assertFalse(self.butler.exists("a_config", {}, collections=[self.butler.run]))
        self.assertFalse(self.butler.exists(PipelineDatasetTypes.packagesDatasetName, {}))

    def test_pre_execute_qgraph_init_outputs(self):
        # Too hard to make a quantum graph from scratch.
        executor = SeparablePipelineExecutor(self.butler)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)

        butlerTests.addDatasetType(self.butler, "output", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "b_config", set(), "Config")
        butlerTests.addDatasetType(self.butler, "b_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "b_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, PipelineDatasetTypes.packagesDatasetName, set(), "Packages")

        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=False,
            save_init_outputs=True,
            save_versions=False,
        )
        self.assertTrue(self.butler.exists("a_config", {}, collections=[self.butler.run]))
        self.assertFalse(self.butler.exists(PipelineDatasetTypes.packagesDatasetName, {}))

    def test_pre_execute_qgraph_versions(self):
        executor = SeparablePipelineExecutor(self.butler)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)

        butlerTests.addDatasetType(self.butler, "output", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "b_config", set(), "Config")
        butlerTests.addDatasetType(self.butler, "b_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "b_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, PipelineDatasetTypes.packagesDatasetName, set(), "Packages")

        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=False,
            save_init_outputs=True,
            save_versions=True,
        )
        self.assertTrue(self.butler.exists("a_config", {}, collections=[self.butler.run]))
        self.assertTrue(self.butler.exists(PipelineDatasetTypes.packagesDatasetName, {}))

    def test_init_badinput(self):
        butler = lsst.daf.butler.Butler(butler=self.butler, collections=[], run="foo")

        with self.assertRaises(ValueError):
            SeparablePipelineExecutor(butler)

    def test_init_badoutput(self):
        butler = lsst.daf.butler.Butler(butler=self.butler, collections=["foo"])

        with self.assertRaises(ValueError):
            SeparablePipelineExecutor(butler)

    def test_make_pipeline_full(self):
        executor = SeparablePipelineExecutor(self.butler)
        for uri in [
            self.pipeline_file,
            ResourcePath(self.pipeline_file),
            ResourcePath(self.pipeline_file).geturl(),
        ]:
            pipeline = executor.make_pipeline(uri)
            self.assertEqual(len(pipeline), 2)
            self.assertEqual({t.label for t in pipeline}, {"a", "b"})

    def test_make_pipeline_subset(self):
        executor = SeparablePipelineExecutor(self.butler)
        path = self.pipeline_file + "#a"
        for uri in [
            path,
            ResourcePath(path),
            ResourcePath(path).geturl(),
        ]:
            pipeline = executor.make_pipeline(uri)
            self.assertEqual(len(pipeline), 1)
            self.assertEqual({t.label for t in pipeline}, {"a"})

    def test_make_quantum_graph_nowhere_noskip_noclobber(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=False)
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 2)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"a"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_noskip_noclobber_conflict(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=False)
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")

        with self.assertRaises(lsst.pipe.base.graphBuilder.OutputExistsError):
            executor.make_quantum_graph(pipeline)

    # TODO: need more complex task and Butler to test
    # make_quantum_graph(where=...)

    def test_make_quantum_graph_nowhere_skipnone_noclobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=False,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 2)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"a"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_skiptotal_noclobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=False,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")
        self.butler.put(lsst.pex.config.Config(), "a_config")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 1)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"b"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_skippartial_noclobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=False,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")

        with self.assertRaises(lsst.pipe.base.graphBuilder.OutputExistsError):
            executor.make_quantum_graph(pipeline)

    def test_make_quantum_graph_nowhere_noskip_clobber(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=True)
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 2)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"a"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_noskip_clobber_conflict(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=True)
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 2)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"a"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_skipnone_clobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 2)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"a"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_skiptotal_clobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")
        self.butler.put(lsst.pex.config.Config(), "a_config")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 1)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"b"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_nowhere_skippartial_clobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")

        graph = executor.make_quantum_graph(pipeline)
        self.assertTrue(graph.isConnected)
        self.assertEqual(len(graph), 2)
        self.assertEqual({q.taskDef.label for q in graph.inputQuanta}, {"a"})
        self.assertEqual({q.taskDef.label for q in graph.outputQuanta}, {"b"})

    def test_make_quantum_graph_noinput(self):
        executor = SeparablePipelineExecutor(self.butler)
        pipeline = Pipeline.fromFile(self.pipeline_file)

        graph = executor.make_quantum_graph(pipeline)
        self.assertEqual(len(graph), 0)

    def test_make_quantum_graph_alloutput_skip(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=[self.butler.run])
        pipeline = Pipeline.fromFile(self.pipeline_file)

        butlerTests.addDatasetType(self.butler, "output", set(), "StructuredDataDict")
        butlerTests.addDatasetType(self.butler, "b_log", set(), "ButlerLogRecords")
        butlerTests.addDatasetType(self.butler, "b_metadata", set(), "TaskMetadata")
        butlerTests.addDatasetType(self.butler, "b_config", set(), "Config")

        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")
        self.butler.put({"zero": 0}, "output")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "b_log")
        self.butler.put(TaskMetadata(), "b_metadata")
        self.butler.put(lsst.pex.config.Config(), "a_config")
        self.butler.put(lsst.pex.config.Config(), "b_config")

        graph = executor.make_quantum_graph(pipeline)
        self.assertEqual(len(graph), 0)

    def test_run_pipeline_noskip_noclobber_fullgraph(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=False)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_run_pipeline_noskip_noclobber_emptygraph(self):
        old_repo_size = self.butler.registry.queryDatasets(...).count()

        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=False)
        graph = lsst.pipe.base.QuantumGraph({}, universe=self.butler.dimensions)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        # Empty graph execution should do nothing.
        self.assertEqual(self.butler.registry.queryDatasets(...).count(), old_repo_size)

    def test_run_pipeline_skipnone_noclobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=False,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_run_pipeline_skiptotal_noclobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=False,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")
        self.butler.put(lsst.pex.config.Config(), "a_config")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("output"), {"zero": 0, "two": 2})

    def test_run_pipeline_noskip_clobber_connected(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=True)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_run_pipeline_noskip_clobber_unconnected(self):
        executor = SeparablePipelineExecutor(self.butler, skip_existing_in=None, clobber_output=True)
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        # The value of output is undefined; it depends on which task ran first.
        self.assertTrue(self.butler.exists("output", {}))

    def test_run_pipeline_skipnone_clobber(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        self.assertEqual(self.butler.get("output"), {"zero": 0, "one": 1, "two": 2})

    def test_run_pipeline_skiptotal_clobber_connected(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        self.butler.put(lsst.daf.butler.ButlerLogRecords.from_records([]), "a_log")
        self.butler.put(TaskMetadata(), "a_metadata")
        self.butler.put(lsst.pex.config.Config(), "a_config")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("output"), {"zero": 0, "two": 2})

    def test_run_pipeline_skippartial_clobber_unconnected(self):
        executor = SeparablePipelineExecutor(
            self.butler,
            skip_existing_in=[self.butler.run],
            clobber_output=True,
        )
        pipeline = Pipeline.fromFile(self.pipeline_file)
        self.butler.put({"zero": 0}, "input")
        self.butler.put({"zero": 0}, "intermediate")
        graph = executor.make_quantum_graph(pipeline)
        executor.pre_execute_qgraph(
            graph,
            register_dataset_types=True,
            save_init_outputs=True,
            save_versions=False,
        )

        executor.run_pipeline(graph)
        self.butler.registry.refresh()
        self.assertEqual(self.butler.get("intermediate"), {"zero": 0, "one": 1})
        # The value of output is undefined; it depends on which task ran first.
        self.assertTrue(self.butler.exists("output", {}))


class MemoryTester(lsst.utils.tests.MemoryTestCase):
    """Generic test for file leaks."""


def setup_module(module):
    """Set up the module for pytest."""
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


__all__ = (
    "butler_options",
    "coverage_options",
    "execution_options",
    "meta_info_options",
    "pipeline_build_options",
    "qgraph_options",
    "run_options",
)


import click
import lsst.daf.butler.cli.opt as dafButlerOpts
import lsst.pipe.base.cli.opt as pipeBaseOpts
from lsst.daf.butler.cli.opt import transfer_option
from lsst.daf.butler.cli.utils import OptionGroup, option_section, unwrap

from . import options as ctrlMpExecOpts

instrumentOptionHelp = (
    "Add an instrument which will be used to load config overrides when "
    "defining a pipeline. This must be the fully qualified class name."
)


class pipeline_build_options(OptionGroup):  # noqa: N801
    """Decorator to add options to the command function for building a
    pipeline.
    """

    def __init__(self) -> None:
        self.decorators = [
            option_section(sectionText="Pipeline build options:"),
            ctrlMpExecOpts.pipeline_option(),
            ctrlMpExecOpts.task_option(),
            ctrlMpExecOpts.delete_option(metavar="LABEL"),
            dafButlerOpts.config_option(metavar="LABEL:NAME=VALUE", multiple=True),
            dafButlerOpts.config_file_option(
                help=unwrap(
                    """Configuration override file(s), applies to a task
                                                         with a given label."""
                ),
                metavar="LABEL:FILE",
                multiple=True,
            ),
            ctrlMpExecOpts.order_pipeline_option(),
            ctrlMpExecOpts.save_pipeline_option(),
            ctrlMpExecOpts.pipeline_dot_option(),
            pipeBaseOpts.instrument_option(help=instrumentOptionHelp, metavar="instrument", multiple=True),
            ctrlMpExecOpts.butler_config_option(required=False),
        ]


class coverage_options(OptionGroup):  # noqa: N801
    """Decorator to add options to the command function for test coverage."""

    def __init__(self) -> None:
        self.decorators = [
            option_section(sectionText="Coverage options:"),
            ctrlMpExecOpts.coverage_option(),
            ctrlMpExecOpts.coverage_report_option(),
            ctrlMpExecOpts.coverage_packages_option(),
        ]


class qgraph_options(OptionGroup):  # noqa: N801
    """Decorator to add options to a command function for creating a quantum
    graph.
    """

    def __init__(self) -> None:
        self.decorators = [
            option_section(sectionText="Quantum graph building options:"),
            ctrlMpExecOpts.qgraph_option(),
            ctrlMpExecOpts.qgraph_id_option(),
            ctrlMpExecOpts.qgraph_node_id_option(),
            ctrlMpExecOpts.qgraph_datastore_records_option(),
            ctrlMpExecOpts.skip_existing_in_option(),
            ctrlMpExecOpts.skip_existing_option(),
            ctrlMpExecOpts.clobber_outputs_option(),
            ctrlMpExecOpts.save_qgraph_option(),
            ctrlMpExecOpts.save_single_quanta_option(),
            ctrlMpExecOpts.qgraph_dot_option(),
            ctrlMpExecOpts.save_execution_butler_option(),
            ctrlMpExecOpts.clobber_execution_butler_option(),
            ctrlMpExecOpts.target_datastore_root_option(),
            transfer_option(
                help=unwrap(
                    """Data transfer mode for the execution butler datastore.
                    Defaults to "copy" if --target-datastore-root is provided.
                    """
                ),
            ),
            ctrlMpExecOpts.dataset_query_constraint(),
            ctrlMpExecOpts.qgraph_header_data_option(),
            ctrlMpExecOpts.mock_option(),
            ctrlMpExecOpts.mock_failure_option(),
            ctrlMpExecOpts.unmocked_dataset_types_option(),
            coverage_options(),
        ]


class butler_options(OptionGroup):  # noqa: N801
    """Decorator to add options to a command function for configuring a
    butler.
    """

    def __init__(self) -> None:
        self.decorators = [
            option_section(sectionText="Data repository and selection options:"),
            ctrlMpExecOpts.butler_config_option(required=True),
            ctrlMpExecOpts.input_option(),
            ctrlMpExecOpts.output_option(),
            ctrlMpExecOpts.output_run_option(),
            ctrlMpExecOpts.extend_run_option(),
            ctrlMpExecOpts.replace_run_option(),
            ctrlMpExecOpts.prune_replaced_option(),
            ctrlMpExecOpts.data_query_option(),
            ctrlMpExecOpts.rebase_option(),
        ]


class execution_options(OptionGroup):  # noqa: N801
    """Decorator to add options to a command function for executing a
    pipeline.
    """

    def __init__(self) -> None:
        self.decorators = [
            option_section(sectionText="Execution options:"),
            ctrlMpExecOpts.clobber_outputs_option(),
            ctrlMpExecOpts.pdb_option(),
            ctrlMpExecOpts.profile_option(),
            dafButlerOpts.processes_option(),
            ctrlMpExecOpts.start_method_option(),
            ctrlMpExecOpts.timeout_option(),
            ctrlMpExecOpts.fail_fast_option(),
            ctrlMpExecOpts.graph_fixup_option(),
            ctrlMpExecOpts.summary_option(),
            ctrlMpExecOpts.enable_implicit_threading_option(),
            ctrlMpExecOpts.cores_per_quantum_option(),
            ctrlMpExecOpts.memory_per_quantum_option(),
        ]


class meta_info_options(OptionGroup):  # noqa: N801
    """Decorator to add options to a command function for managing pipeline
    meta information.
    """

    def __init__(self) -> None:
        self.decorators = [
            option_section(sectionText="Meta-information output options:"),
            ctrlMpExecOpts.skip_init_writes_option(),
            ctrlMpExecOpts.init_only_option(),
            dafButlerOpts.register_dataset_types_option(),
            ctrlMpExecOpts.no_versions_option(),
        ]


class run_options(OptionGroup):  # noqa: N801
    """Decorator to add the run options to the run command."""

    def __init__(self) -> None:
        self.decorators = [
            click.pass_context,
            ctrlMpExecOpts.debug_option(),
            ctrlMpExecOpts.show_option(),
            pipeline_build_options(),
            qgraph_options(),
            butler_options(),
            execution_options(),
            meta_info_options(),
            coverage_options(),
            option_section(sectionText=""),
            dafButlerOpts.options_file_option(),
        ]



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simple unit test for PreExecInit class.
"""

import contextlib
import shutil
import tempfile
import unittest

from lsst.ctrl.mpexec import PreExecInit
from lsst.pipe.base.tests.simpleQGraph import AddTaskFactoryMock, makeSimpleQGraph


@contextlib.contextmanager
def temporaryDirectory():
    """Context manager that creates and destroys temporary directory.

    Difference from `tempfile.TemporaryDirectory` is that it ignores errors
    when deleting a directory, which may happen with some filesystems.
    """
    tmpdir = tempfile.mkdtemp()
    yield tmpdir
    shutil.rmtree(tmpdir, ignore_errors=True)


class PreExecInitTestCase(unittest.TestCase):
    """A test case for PreExecInit"""

    def test_saveInitOutputs(self):
        taskFactory = AddTaskFactoryMock()
        for extendRun in (False, True):
            with self.subTest(extendRun=extendRun):
                with temporaryDirectory() as tmpdir:
                    butler, qgraph = makeSimpleQGraph(root=tmpdir)
                    preExecInit = PreExecInit(butler=butler, taskFactory=taskFactory, extendRun=extendRun)
                    preExecInit.saveInitOutputs(qgraph)

    def test_saveInitOutputs_twice(self):
        taskFactory = AddTaskFactoryMock()
        for extendRun in (False, True):
            with self.subTest(extendRun=extendRun):
                with temporaryDirectory() as tmpdir:
                    butler, qgraph = makeSimpleQGraph(root=tmpdir)
                    preExecInit = PreExecInit(butler=butler, taskFactory=taskFactory, extendRun=extendRun)
                    preExecInit.saveInitOutputs(qgraph)
                    if extendRun:
                        # will ignore this
                        preExecInit.saveInitOutputs(qgraph)
                    else:
                        # Second time it will fail
                        with self.assertRaises(Exception):
                            preExecInit.saveInitOutputs(qgraph)

    def test_saveConfigs(self):
        for extendRun in (False, True):
            with self.subTest(extendRun=extendRun):
                with temporaryDirectory() as tmpdir:
                    butler, qgraph = makeSimpleQGraph(root=tmpdir)
                    preExecInit = PreExecInit(butler=butler, taskFactory=None, extendRun=extendRun)
                    preExecInit.saveConfigs(qgraph)

    def test_saveConfigs_twice(self):
        for extendRun in (False, True):
            with self.subTest(extendRun=extendRun):
                with temporaryDirectory() as tmpdir:
                    butler, qgraph = makeSimpleQGraph(root=tmpdir)
                    preExecInit = PreExecInit(butler=butler, taskFactory=None, extendRun=extendRun)
                    preExecInit.saveConfigs(qgraph)
                    if extendRun:
                        # will ignore this
                        preExecInit.saveConfigs(qgraph)
                    else:
                        # Second time it will fail
                        with self.assertRaises(Exception):
                            preExecInit.saveConfigs(qgraph)

    def test_savePackageVersions(self):
        for extendRun in (False, True):
            with self.subTest(extendRun=extendRun):
                with temporaryDirectory() as tmpdir:
                    butler, qgraph = makeSimpleQGraph(root=tmpdir)
                    preExecInit = PreExecInit(butler=butler, taskFactory=None, extendRun=extendRun)
                    preExecInit.savePackageVersions(qgraph)

    def test_savePackageVersions_twice(self):
        for extendRun in (False, True):
            with self.subTest(extendRun=extendRun):
                with temporaryDirectory() as tmpdir:
                    butler, qgraph = makeSimpleQGraph(root=tmpdir)
                    preExecInit = PreExecInit(butler=butler, taskFactory=None, extendRun=extendRun)
                    preExecInit.savePackageVersions(qgraph)
                    if extendRun:
                        # If this is the same packages then it should not
                        # attempt to save.
                        preExecInit.savePackageVersions(qgraph)
                    else:
                        # second time it will fail
                        with self.assertRaises(Exception):
                            preExecInit.savePackageVersions(qgraph)


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for ctrl_mpexec CLI purge subcommand.
"""


import os
import unittest

from lsst.ctrl.mpexec.cli.pipetask import cli as pipetask_cli
from lsst.daf.butler.cli.butler import cli as butler_cli
from lsst.daf.butler.cli.utils import LogCliRunner, clickResultMsg
from lsst.daf.butler.tests.utils import MetricTestRepo, makeTestTempDir, removeTestTempDir

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class PurgeTest(unittest.TestCase):
    """Test executing "pipetask purge" commands."""

    def setUp(self):
        self.runner = LogCliRunner()

        # this creates a repo with collections:
        #    Name     Type
        # ---------- ------
        # ingest     TAGGED
        # ingest/run RUN
        self.root = makeTestTempDir(TESTDIR)
        self.testRepo = MetricTestRepo(
            self.root,
            configFile=os.path.join(TESTDIR, "config/metricTestRepoButler.yaml"),
        )

    def tearDown(self):
        removeTestTempDir(self.root)

    def test_singleChain_yesNo(self):
        """Test removing a chain with one child, and the yes/no
        confirmation.
        """
        # add the collection ingest/run to a CHAINED collection called "in"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "in", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # purge the CHAINED collection called "in", but say "no", check for
        # expected outputs.
        result = self.runner.invoke(pipetask_cli, ["purge", "-b", self.root, "in"], input="no")
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        self.assertIn("Will remove:\n  runs: ingest/run\n  chains: in\n  others: \n", result.output)
        self.assertIn("Aborted.", result.output)

        # purge the CHAINED collection called "in", and say "yes", check for
        # expected outputs.
        result = self.runner.invoke(pipetask_cli, ["purge", "-b", self.root, "in"], input="yes")
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        self.assertIn("Will remove:\n  runs: ingest/run\n  chains: in\n  others: \n", result.output)
        self.assertIn("Done.", result.output)

    def test_granparentChain_noConfirm(self):
        """Test removing a chain with children and grandchildren, and the
        --no-confirm option.
        """
        # add the collection ingest/run to a CHAINED collection called "ing"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "ing", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # add the CHAINED collectin "ing" a CHAINED collection called "in"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "in", "ing"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # purge the CHAINED collection called "in" with --no-confirm and check
        # for expected outputs.
        result = self.runner.invoke(
            pipetask_cli,
            ["purge", "-b", self.root, "in", "--recursive", "--no-confirm"],
        )
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        self.assertIn("Removed:\n  runs: ingest/run\n  chains: in, ing\n  others: \n", result.output)

    def test_topParentWithParent(self):
        """Test that purging a chain with a parent fails."""
        # add the collection ingest/run to a CHAINED collection called "ing"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "ing", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # add the CHAINED collectin "ing" a CHAINED collection called "in"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "in", "ing"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # purge the CHAINED collection called "ing" and check for expected
        # outputs.
        result = self.runner.invoke(
            pipetask_cli,
            ["purge", "-b", self.root, "ing"],
        )
        self.assertEqual(result.exit_code, 1, clickResultMsg(result))
        self.assertIn(
            'The passed-in collection "ing" must not be contained in other collections but '
            'is contained in collection(s) "in".',
            result.output,
        )

    def test_childWithMultipleParents(self):
        """Test that a child chain with multiple parents fails."""
        # add the collection ingest/run to a CHAINED collection called "ing"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "ing", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # add the collectin ingest/run to a CHAINED collection called "foo"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "foo", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # purge the CHAINED collection called "ing" and check for expected
        # outputs.
        result = self.runner.invoke(
            pipetask_cli,
            ["purge", "-b", self.root, "ing"],
        )
        self.assertEqual(result.exit_code, 1, clickResultMsg(result))
        self.assertIn(
            'Collection "ingest/run" is in multiple chained collections:',
            result.output,
        )
        self.assertIn('"foo"', result.output)
        self.assertIn('"ing"', result.output)

    def test_notFound_notChained(self):
        """Test for failure when the top level collection is not found,
        and when a top level connection is not a CHAINED collection.
        """
        # Test purging a collection that does not exist.
        result = self.runner.invoke(
            pipetask_cli,
            ["purge", "-b", self.root, "notACollection"],
        )
        self.assertEqual(result.exit_code, 1, clickResultMsg(result))
        self.assertIn(
            'The passed-in collection "notACollection" was not found.',
            result.output,
        )

        # Test purging a collection that is not a CHAINED collection.
        result = self.runner.invoke(
            pipetask_cli,
            ["purge", "-b", self.root, "ingest/run"],
        )
        self.assertEqual(result.exit_code, 1, clickResultMsg(result))
        self.assertIn(
            'The passed-in collection must be a CHAINED collection; "ingest/run" is a RUN collection.',
            result.output,
        )


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from types import SimpleNamespace

from ... import CmdLineFwk, TaskFactory


def pre_exec_init_qbb(
    butler_config: str,
    qgraph: str,
    qgraph_id: str | None,
    config_search_path: list[str] | None,
) -> None:
    """Implement the command line interface ``pipetask pre-exec-init-qbb``
    subcommand.

    Should only be called by command line tools and unit test code
    that tests this function.

    Parameters
    ----------
    butler_config : `str`
        The path location of the gen3 butler/registry config file.
    qgraph : `str`
        URI location for a serialized quantum graph definition.
    qgraph_id : `str` or `None`
        Quantum graph identifier, if specified must match the identifier of the
        graph loaded from a file. Ignored if graph is not loaded from a file.
    config_search_path : `list` [`str`]
        Additional search paths for butler configuration.
    """
    args = SimpleNamespace(
        butler_config=butler_config,
        qgraph=qgraph,
        qgraph_id=qgraph_id,
        config_search_path=config_search_path,
    )

    f = CmdLineFwk()
    task_factory = TaskFactory()
    f.preExecInitQBB(task_factory, args)



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["SingleQuantumExecutor"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
import logging
import sys
import time
from collections import defaultdict
from collections.abc import Callable
from itertools import chain
from typing import Any

from lsst.daf.butler import (
    Butler,
    CollectionType,
    DatasetRef,
    DatasetType,
    LimitedButler,
    NamedKeyDict,
    Quantum,
)
from lsst.daf.butler.registry.wildcards import CollectionWildcard
from lsst.pipe.base import (
    AdjustQuantumHelper,
    ExecutionResources,
    Instrument,
    InvalidQuantumError,
    NoWorkFound,
    PipelineTask,
    QuantumContext,
    RepeatableQuantumError,
    TaskDef,
    TaskFactory,
)

# During metadata transition phase, determine metadata class by
# asking pipe_base
from lsst.pipe.base.task import _TASK_FULL_METADATA_TYPE, _TASK_METADATA_TYPE
from lsst.utils.timer import logInfo

# -----------------------------
#  Imports for other modules --
# -----------------------------
from .log_capture import LogCapture
from .quantumGraphExecutor import QuantumExecutor
from .reports import QuantumReport

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

_LOG = logging.getLogger(__name__)


class SingleQuantumExecutor(QuantumExecutor):
    """Executor class which runs one Quantum at a time.

    Parameters
    ----------
    butler : `~lsst.daf.butler.Butler` or `None`
        Data butler, `None` means that Quantum-backed butler should be used
        instead.
    taskFactory : `~lsst.pipe.base.TaskFactory`
        Instance of a task factory.
    skipExistingIn
        Expressions representing the collections to search for existing
        output datasets. See :ref:`daf_butler_ordered_collection_searches`
        for allowed types. This class only checks for the presence of butler
        output run in the list of collections. If the output run is present
        in the list then the quanta whose complete outputs exist in the output
        run will be skipped. `None` or empty string/sequence disables skipping.
    clobberOutputs : `bool`, optional
        If `True`, then outputs from a quantum that exist in output run
        collection will be removed prior to executing a quantum. If
        ``skipExistingIn`` contains output run, then only partial outputs from
        a quantum will be removed. Only used when ``butler`` is not `None`.
    enableLsstDebug : `bool`, optional
        Enable debugging with ``lsstDebug`` facility for a task.
    exitOnKnownError : `bool`, optional
        If `True`, call `sys.exit` with the appropriate exit code for special
        known exceptions, after printing a traceback, instead of letting the
        exception propagate up to calling.  This is always the behavior for
        InvalidQuantumError.
    limited_butler_factory : `Callable`, optional
        A method that creates a `~lsst.daf.butler.LimitedButler` instance
        for a given Quantum. This parameter must be defined if ``butler`` is
        `None`. If ``butler`` is not `None` then this parameter is ignored.
    resources : `~lsst.pipe.base.ExecutionResources`, optional
        The resources available to this quantum when executing.
    """

    def __init__(
        self,
        butler: Butler | None,
        taskFactory: TaskFactory,
        skipExistingIn: Any = None,
        clobberOutputs: bool = False,
        enableLsstDebug: bool = False,
        exitOnKnownError: bool = False,
        limited_butler_factory: Callable[[Quantum], LimitedButler] | None = None,
        resources: ExecutionResources | None = None,
    ):
        self.butler = butler
        self.taskFactory = taskFactory
        self.enableLsstDebug = enableLsstDebug
        self.clobberOutputs = clobberOutputs
        self.exitOnKnownError = exitOnKnownError
        self.limited_butler_factory = limited_butler_factory
        self.report: QuantumReport | None = None
        self.resources = resources

        if self.butler is None:
            assert limited_butler_factory is not None, "limited_butler_factory is needed when butler is None"

        # Find whether output run is in skipExistingIn.
        # TODO: This duplicates logic in GraphBuilder, would be nice to have
        # better abstraction for this some day.
        self.skipExisting = False
        if self.butler is not None and skipExistingIn:
            skip_collections_wildcard = CollectionWildcard.from_expression(skipExistingIn)
            # As optimization check in the explicit list of names first
            self.skipExisting = self.butler.run in skip_collections_wildcard.strings
            if not self.skipExisting:
                # need to flatten it and check again
                self.skipExisting = self.butler.run in self.butler.registry.queryCollections(
                    skipExistingIn,
                    collectionTypes=CollectionType.RUN,
                )

    def execute(self, taskDef: TaskDef, quantum: Quantum) -> Quantum:
        # Docstring inherited from QuantumExecutor.execute
        assert quantum.dataId is not None, "Quantum DataId cannot be None"

        if self.butler is not None:
            self.butler.registry.refresh()

        # Catch any exception and make a report based on that.
        try:
            result = self._execute(taskDef, quantum)
            self.report = QuantumReport(dataId=quantum.dataId, taskLabel=taskDef.label)
            return result
        except Exception as exc:
            self.report = QuantumReport.from_exception(
                exception=exc,
                dataId=quantum.dataId,
                taskLabel=taskDef.label,
            )
            raise

    def _execute(self, taskDef: TaskDef, quantum: Quantum) -> Quantum:
        """Execute the quantum.

        Internal implementation of `execute()`.
        """
        startTime = time.time()

        # Make a limited butler instance if needed (which should be QBB if full
        # butler is not defined).
        limited_butler: LimitedButler
        if self.butler is not None:
            limited_butler = self.butler
        else:
            # We check this in constructor, but mypy needs this check here.
            assert self.limited_butler_factory is not None
            limited_butler = self.limited_butler_factory(quantum)

        if self.butler is not None:
            log_capture = LogCapture.from_full(self.butler)
        else:
            log_capture = LogCapture.from_limited(limited_butler)
        with log_capture.capture_logging(taskDef, quantum) as captureLog:
            # Save detailed resource usage before task start to metadata.
            quantumMetadata = _TASK_METADATA_TYPE()
            logInfo(None, "prep", metadata=quantumMetadata)  # type: ignore[arg-type]

            _LOG.info("Preparing execution of quantum for label=%s dataId=%s.", taskDef.label, quantum.dataId)

            # check whether to skip or delete old outputs, if it returns True
            # or raises an exception do not try to store logs, as they may be
            # already in butler.
            captureLog.store = False
            if self.checkExistingOutputs(quantum, taskDef, limited_butler):
                _LOG.info(
                    "Skipping already-successful quantum for label=%s dataId=%s.",
                    taskDef.label,
                    quantum.dataId,
                )
                return quantum
            captureLog.store = True

            try:
                quantum = self.updatedQuantumInputs(quantum, taskDef, limited_butler)
            except NoWorkFound as exc:
                _LOG.info(
                    "Nothing to do for task '%s' on quantum %s; saving metadata and skipping: %s",
                    taskDef.label,
                    quantum.dataId,
                    str(exc),
                )
                # Make empty metadata that looks something like what a
                # do-nothing task would write (but we don't bother with empty
                # nested PropertySets for subtasks).  This is slightly
                # duplicative with logic in pipe_base that we can't easily call
                # from here; we'll fix this on DM-29761.
                logInfo(None, "end", metadata=quantumMetadata)  # type: ignore[arg-type]
                fullMetadata = _TASK_FULL_METADATA_TYPE()
                fullMetadata[taskDef.label] = _TASK_METADATA_TYPE()
                fullMetadata["quantum"] = quantumMetadata
                self.writeMetadata(quantum, fullMetadata, taskDef, limited_butler)
                return quantum

            # enable lsstDebug debugging
            if self.enableLsstDebug:
                try:
                    _LOG.debug("Will try to import debug.py")
                    import debug  # type: ignore # noqa:F401
                except ImportError:
                    _LOG.warn("No 'debug' module found.")

            # initialize global state
            self.initGlobals(quantum)

            # Ensure that we are executing a frozen config
            taskDef.config.freeze()
            logInfo(None, "init", metadata=quantumMetadata)  # type: ignore[arg-type]
            init_input_refs = list(quantum.initInputs.values())

            _LOG.info(
                "Constructing task and executing quantum for label=%s dataId=%s.",
                taskDef.label,
                quantum.dataId,
            )
            task = self.taskFactory.makeTask(taskDef, limited_butler, init_input_refs)
            logInfo(None, "start", metadata=quantumMetadata)  # type: ignore[arg-type]
            try:
                self.runQuantum(task, quantum, taskDef, limited_butler)
            except Exception as e:
                _LOG.error(
                    "Execution of task '%s' on quantum %s failed. Exception %s: %s",
                    taskDef.label,
                    quantum.dataId,
                    e.__class__.__name__,
                    str(e),
                )
                raise
            logInfo(None, "end", metadata=quantumMetadata)  # type: ignore[arg-type]
            fullMetadata = task.getFullMetadata()
            fullMetadata["quantum"] = quantumMetadata
            self.writeMetadata(quantum, fullMetadata, taskDef, limited_butler)
            stopTime = time.time()
            _LOG.info(
                "Execution of task '%s' on quantum %s took %.3f seconds",
                taskDef.label,
                quantum.dataId,
                stopTime - startTime,
            )
        return quantum

    def checkExistingOutputs(self, quantum: Quantum, taskDef: TaskDef, limited_butler: LimitedButler) -> bool:
        """Decide whether this quantum needs to be executed.

        If only partial outputs exist then they are removed if
        ``clobberOutputs`` is True, otherwise an exception is raised.

        Parameters
        ----------
        quantum : `~lsst.daf.butler.Quantum`
            Quantum to check for existing outputs
        taskDef : `~lsst.pipe.base.TaskDef`
            Task definition structure.

        Returns
        -------
        exist : `bool`
            `True` if ``self.skipExisting`` is defined, and a previous
            execution of this quanta appears to have completed successfully
            (either because metadata was written or all datasets were written).
            `False` otherwise.

        Raises
        ------
        RuntimeError
            Raised if some outputs exist and some not.
        """
        if not self.butler:
            # Skip/prune logic only works for full butler.
            return False

        if self.skipExisting:
            _LOG.debug(
                "Checking existence of metadata from previous execution of label=%s dataId=%s.",
                taskDef.label,
                quantum.dataId,
            )
            # Metadata output exists; this is sufficient to assume the previous
            # run was successful and should be skipped.
            [metadata_ref] = quantum.outputs[taskDef.metadataDatasetName]
            if metadata_ref is not None:
                if limited_butler.stored(metadata_ref):
                    return True

        # Find and prune (partial) outputs if `self.clobberOutputs` is set.
        _LOG.debug(
            "Looking for existing outputs in the way for label=%s dataId=%s.", taskDef.label, quantum.dataId
        )
        ref_dict = self.butler.stored_many(chain.from_iterable(quantum.outputs.values()))
        existingRefs = [ref for ref, exists in ref_dict.items() if exists]
        missingRefs = [ref for ref, exists in ref_dict.items() if not exists]
        if existingRefs:
            if not missingRefs:
                # Full outputs exist.
                if self.skipExisting:
                    return True
                elif self.clobberOutputs:
                    _LOG.info("Removing complete outputs for quantum %s: %s", quantum, existingRefs)
                    self.butler.pruneDatasets(existingRefs, disassociate=True, unstore=True, purge=True)
                else:
                    raise RuntimeError(
                        f"Complete outputs exists for a quantum {quantum} "
                        "and neither clobberOutputs nor skipExisting is set: "
                        f"collection={self.butler.run} existingRefs={existingRefs}"
                    )
            else:
                # Partial outputs from a failed quantum.
                _LOG.debug(
                    "Partial outputs exist for quantum %s collection=%s existingRefs=%s missingRefs=%s",
                    quantum,
                    self.butler.run,
                    existingRefs,
                    missingRefs,
                )
                if self.clobberOutputs:
                    # only prune
                    _LOG.info("Removing partial outputs for task %s: %s", taskDef, existingRefs)
                    self.butler.pruneDatasets(existingRefs, disassociate=True, unstore=True, purge=True)
                    return False
                else:
                    raise RuntimeError(
                        "Registry inconsistency while checking for existing quantum outputs:"
                        f" quantum={quantum} collection={self.butler.run} existingRefs={existingRefs}"
                        f" missingRefs={missingRefs}"
                    )

        # By default always execute.
        return False

    def updatedQuantumInputs(
        self, quantum: Quantum, taskDef: TaskDef, limited_butler: LimitedButler
    ) -> Quantum:
        """Update quantum with extra information, returns a new updated
        Quantum.

        Some methods may require input DatasetRefs to have non-None
        ``dataset_id``, but in case of intermediate dataset it may not be
        filled during QuantumGraph construction. This method will retrieve
        missing info from registry.

        Parameters
        ----------
        quantum : `~lsst.daf.butler.Quantum`
            Single Quantum instance.
        taskDef : `~lsst.pipe.base.TaskDef`
            Task definition structure.

        Returns
        -------
        update : `~lsst.daf.butler.Quantum`
            Updated Quantum instance
        """
        anyChanges = False
        updatedInputs: defaultdict[DatasetType, list] = defaultdict(list)
        for key, refsForDatasetType in quantum.inputs.items():
            _LOG.debug(
                "Checking existence of input '%s' for label=%s dataId=%s.",
                key.name,
                taskDef.label,
                quantum.dataId,
            )
            newRefsForDatasetType = updatedInputs[key]
            stored = limited_butler.stored_many(refsForDatasetType)
            for ref in refsForDatasetType:
                if stored[ref]:
                    newRefsForDatasetType.append(ref)
                else:
                    # This should only happen if a predicted intermediate was
                    # not actually produced upstream, but
                    # datastore misconfigurations can unfortunately also land
                    # us here.
                    _LOG.info("No dataset artifact found for %s", ref)
                    continue
            if len(newRefsForDatasetType) != len(refsForDatasetType):
                anyChanges = True
        # If we removed any input datasets, let the task check if it has enough
        # to proceed and/or prune related datasets that it also doesn't
        # need/produce anymore.  It will raise NoWorkFound if it can't run,
        # which we'll let propagate up.  This is exactly what we run during QG
        # generation, because a task shouldn't care whether an input is missing
        # because some previous task didn't produce it, or because it just
        # wasn't there during QG generation.
        namedUpdatedInputs = NamedKeyDict[DatasetType, list[DatasetRef]](updatedInputs.items())
        helper = AdjustQuantumHelper(namedUpdatedInputs, quantum.outputs)
        if anyChanges:
            _LOG.debug("Running adjustQuantum for label=%s dataId=%s.", taskDef.label, quantum.dataId)
            assert quantum.dataId is not None, "Quantum DataId cannot be None"
            helper.adjust_in_place(taskDef.connections, label=taskDef.label, data_id=quantum.dataId)
        return Quantum(
            taskName=quantum.taskName,
            taskClass=quantum.taskClass,
            dataId=quantum.dataId,
            initInputs=quantum.initInputs,
            inputs=helper.inputs,
            outputs=helper.outputs,
        )

    def runQuantum(
        self, task: PipelineTask, quantum: Quantum, taskDef: TaskDef, limited_butler: LimitedButler
    ) -> None:
        """Execute task on a single quantum.

        Parameters
        ----------
        task : `~lsst.pipe.base.PipelineTask`
            Task object.
        quantum : `~lsst.daf.butler.Quantum`
            Single Quantum instance.
        taskDef : `~lsst.pipe.base.TaskDef`
            Task definition structure.
        """
        # Create a butler that operates in the context of a quantum
        butlerQC = QuantumContext(limited_butler, quantum, resources=self.resources)

        # Get the input and output references for the task
        inputRefs, outputRefs = taskDef.connections.buildDatasetRefs(quantum)

        # Call task runQuantum() method.  Catch a few known failure modes and
        # translate them into specific
        try:
            task.runQuantum(butlerQC, inputRefs, outputRefs)
        except NoWorkFound as err:
            # Not an error, just an early exit.
            _LOG.info("Task '%s' on quantum %s exited early: %s", taskDef.label, quantum.dataId, str(err))
            pass
        except RepeatableQuantumError as err:
            if self.exitOnKnownError:
                _LOG.warning("Caught repeatable quantum error for %s (%s):", taskDef, quantum.dataId)
                _LOG.warning(err, exc_info=True)
                sys.exit(err.EXIT_CODE)
            else:
                raise
        except InvalidQuantumError as err:
            _LOG.fatal("Invalid quantum error for %s (%s): %s", taskDef, quantum.dataId)
            _LOG.fatal(err, exc_info=True)
            sys.exit(err.EXIT_CODE)

    def writeMetadata(
        self, quantum: Quantum, metadata: Any, taskDef: TaskDef, limited_butler: LimitedButler
    ) -> None:
        # DatasetRef has to be in the Quantum outputs, can lookup by name
        try:
            [ref] = quantum.outputs[taskDef.metadataDatasetName]
        except LookupError as exc:
            raise InvalidQuantumError(
                f"Quantum outputs is missing metadata dataset type {taskDef.metadataDatasetName};"
                " this could happen due to inconsistent options between QuantumGraph generation"
                " and execution"
            ) from exc
        limited_butler.put(metadata, ref)

    def initGlobals(self, quantum: Quantum) -> None:
        """Initialize global state needed for task execution.

        Parameters
        ----------
        quantum : `~lsst.daf.butler.Quantum`
            Single Quantum instance.

        Notes
        -----
        There is an issue with initializing filters singleton which is done
        by instrument, to avoid requiring tasks to do it in runQuantum()
        we do it here when any dataId has an instrument dimension. Also for
        now we only allow single instrument, verify that all instrument
        names in all dataIds are identical.

        This will need revision when filter singleton disappears.
        """
        # can only work for full butler
        if self.butler is None:
            return
        oneInstrument = None
        for datasetRefs in chain(quantum.inputs.values(), quantum.outputs.values()):
            for datasetRef in datasetRefs:
                dataId = datasetRef.dataId
                instrument = dataId.get("instrument")
                if instrument is not None:
                    if oneInstrument is not None:
                        assert (  # type: ignore
                            instrument == oneInstrument
                        ), "Currently require that only one instrument is used per graph"
                    else:
                        oneInstrument = instrument
                        Instrument.fromName(instrument, self.butler.registry)

    def getReport(self) -> QuantumReport | None:
        # Docstring inherited from base class
        if self.report is None:
            raise RuntimeError("getReport() called before execute()")
        return self.report



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for ctrl_mpexec CLI cleanup subcommand.
"""


import os
import unittest

from lsst.ctrl.mpexec.cli.pipetask import cli as pipetask_cli
from lsst.daf.butler import Butler
from lsst.daf.butler.cli.butler import cli as butler_cli
from lsst.daf.butler.cli.utils import LogCliRunner, clickResultMsg
from lsst.daf.butler.tests.utils import MetricTestRepo, makeTestTempDir, removeTestTempDir

TESTDIR = os.path.abspath(os.path.dirname(__file__))


class CleanupCollectionTest(unittest.TestCase):
    """Test executing "pipetask cleanup" commands."""

    def setUp(self):
        self.runner = LogCliRunner()

        # this creates a repo with collections:
        #    Name     Type
        # ---------- ------
        # ingest     TAGGED
        # ingest/run RUN
        self.root = makeTestTempDir(TESTDIR)
        self.testRepo = MetricTestRepo(
            self.root,
            configFile=os.path.join(TESTDIR, "config/metricTestRepoButler.yaml"),
        )

    def tearDown(self):
        removeTestTempDir(self.root)

    def test_cleanup_yesNo(self):
        """Test cleaning up a non-chained collection and that other collections
        in the chain remian. Verify the yes/no dialog works.
        """
        # add the collection ingest/run to a CHAINED collection called "in"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "in", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # clean up the CHAINED collection "in", verify that the TAGGED
        # collection "ingest" is removed, but the child RUN collection
        # "ingest/run" is not removed.
        result = self.runner.invoke(pipetask_cli, ["cleanup", "-b", self.root, "in"], input="n")
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        self.assertIn("Will remove:\n  runs: \n  others: ingest\n", result.output)
        self.assertIn("Aborted.", result.output)

        # run cleanup again but say "y" to continue, and check for expected
        # outputs.
        result = self.runner.invoke(pipetask_cli, ["cleanup", "-b", self.root, "in"], input="y")
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        self.assertIn("Will remove:\n  runs: \n  others: ingest\n", result.output)
        self.assertIn("Done.", result.output)

        butler = Butler(self.root)
        self.assertEqual(set(butler.registry.queryCollections()), {"in", "ingest/run"})

    def test_nonExistantCollection(self):
        """Test running cleanup on a collection that has never existed."""
        result = self.runner.invoke(pipetask_cli, ["cleanup", "-b", self.root, "foo"])
        self.assertEqual(result.exit_code, 1, clickResultMsg(result))
        self.assertIn('Did not find a collection named "foo"', result.output)

    def test_removedCollection(self):
        """Test cleaning up a collection that used to exist."""
        # Remove the TAGGED collection called "ingest"
        result = self.runner.invoke(butler_cli, ["remove-collections", self.root, "ingest"], input="y")
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # Add the collection ingest/run to a CHAINED collection called "in"
        result = self.runner.invoke(butler_cli, ["collection-chain", self.root, "in", "ingest/run"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        # Purge the collection
        result = self.runner.invoke(pipetask_cli, ["purge", "-b", self.root, "in"], input="no")
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))

        # Attempt to clean up the collection, expect a message that there was
        # nothing to do.
        result = self.runner.invoke(pipetask_cli, ["cleanup", "-b", self.root, "in"])
        self.assertEqual(result.exit_code, 0, clickResultMsg(result))
        self.assertIn("Did not find any collections to remove.", result.output)

    def test_cleanupNonChained(self):
        result = self.runner.invoke(pipetask_cli, ["cleanup", "-b", self.root, "ingest"])
        self.assertEqual(result.exit_code, 1, clickResultMsg(result))
        self.assertIn(
            'Error: COLLECTION must be a CHAINED collection, "ingest" is a "TAGGED" collection.',
            result.output,
        )


if __name__ == "__main__":
    unittest.main()



# This file is part of ctrl_mpexec.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["PreExecInit"]

# -------------------------------
#  Imports of standard modules --
# -------------------------------
import abc
import logging
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

# -----------------------------
#  Imports for other modules --
# -----------------------------
from lsst.daf.butler import DatasetRef, DatasetType
from lsst.daf.butler.registry import ConflictingDefinitionError
from lsst.pipe.base import PipelineDatasetTypes
from lsst.utils.packages import Packages

if TYPE_CHECKING:
    from lsst.daf.butler import Butler, LimitedButler
    from lsst.pipe.base import QuantumGraph, TaskDef, TaskFactory

_LOG = logging.getLogger(__name__)


class MissingReferenceError(Exception):
    """Exception raised when resolved reference is missing from graph."""

    pass


def _compare_packages(old_packages: Packages, new_packages: Packages) -> None:
    """Compare two versions of Packages.

    Parameters
    ----------
    old_packages : `Packages`
        Previously recorded package versions.
    new_packages : `Packages`
        New set of package versions.

    Raises
    ------
    TypeError
        Raised if parameters are inconsistent.
    """
    diff = new_packages.difference(old_packages)
    if diff:
        versions_str = "; ".join(f"{pkg}: {diff[pkg][1]} vs {diff[pkg][0]}" for pkg in diff)
        raise TypeError(f"Package versions mismatch: ({versions_str})")
    else:
        _LOG.debug("new packages are consistent with old")


class PreExecInitBase(abc.ABC):
    """Common part of the implementation of PreExecInit classes that does not
    depend on Butler type.
    """

    def __init__(self, butler: LimitedButler, taskFactory: TaskFactory, extendRun: bool):
        self.butler = butler
        self.taskFactory = taskFactory
        self.extendRun = extendRun

    def initialize(
        self,
        graph: QuantumGraph,
        saveInitOutputs: bool = True,
        registerDatasetTypes: bool = False,
        saveVersions: bool = True,
    ) -> None:
        """Perform all initialization steps.

        Convenience method to execute all initialization steps. Instead of
        calling this method and providing all options it is also possible to
        call methods individually.

        Parameters
        ----------
        graph : `~lsst.pipe.base.QuantumGraph`
            Execution graph.
        saveInitOutputs : `bool`, optional
            If ``True`` (default) then save "init outputs", configurations,
            and package versions to butler.
        registerDatasetTypes : `bool`, optional
            If ``True`` then register dataset types in registry, otherwise
            they must be already registered.
        saveVersions : `bool`, optional
            If ``False`` then do not save package versions even if
            ``saveInitOutputs`` is set to ``True``.
        """
        # register dataset types or check consistency
        self.initializeDatasetTypes(graph, registerDatasetTypes)

        # Save task initialization data or check that saved data
        # is consistent with what tasks would save
        if saveInitOutputs:
            self.saveInitOutputs(graph)
            self.saveConfigs(graph)
            if saveVersions:
                self.savePackageVersions(graph)

    @abc.abstractmethod
    def initializeDatasetTypes(self, graph: QuantumGraph, registerDatasetTypes: bool = False) -> None:
        """Save or check DatasetTypes output by the tasks in a graph.

        Iterates over all DatasetTypes for all tasks in a graph and either
        tries to add them to registry or compares them to existing ones.

        Parameters
        ----------
        graph : `~lsst.pipe.base.QuantumGraph`
            Execution graph.
        registerDatasetTypes : `bool`, optional
            If ``True`` then register dataset types in registry, otherwise
            they must be already registered.

        Raises
        ------
        ValueError
            Raised if existing DatasetType is different from DatasetType
            in a graph.
        KeyError
            Raised if ``registerDatasetTypes`` is ``False`` and DatasetType
            does not exist in registry.
        """
        raise NotImplementedError()

    def saveInitOutputs(self, graph: QuantumGraph) -> None:
        """Write any datasets produced by initializing tasks in a graph.

        Parameters
        ----------
        graph : `~lsst.pipe.base.QuantumGraph`
            Execution graph.

        Raises
        ------
        TypeError
            Raised if the type of existing object in butler is different from
            new data.
        """
        _LOG.debug("Will save InitOutputs for all tasks")
        for taskDef in self._task_iter(graph):
            init_input_refs = graph.initInputRefs(taskDef) or []
            task = self.taskFactory.makeTask(taskDef, self.butler, init_input_refs)
            for name in taskDef.connections.initOutputs:
                attribute = getattr(taskDef.connections, name)
                init_output_refs = graph.initOutputRefs(taskDef) or []
                init_output_ref, obj_from_store = self._find_dataset(init_output_refs, attribute.name)
                if init_output_ref is None:
                    raise ValueError(f"Cannot find dataset reference for init output {name} in a graph")
                init_output_var = getattr(task, name)

                if obj_from_store is not None:
                    _LOG.debug(
                        "Retrieving InitOutputs for task=%s key=%s dsTypeName=%s", task, name, attribute.name
                    )
                    obj_from_store = self.butler.get(init_output_ref)
                    # Types are supposed to be identical.
                    # TODO: Check that object contents is identical too.
                    if type(obj_from_store) is not type(init_output_var):
                        raise TypeError(
                            f"Stored initOutput object type {type(obj_from_store)} "
                            "is different from task-generated type "
                            f"{type(init_output_var)} for task {taskDef}"
                        )
                else:
                    _LOG.debug("Saving InitOutputs for task=%s key=%s", taskDef.label, name)
                    # This can still raise if there is a concurrent write.
                    self.butler.put(init_output_var, init_output_ref)

    def saveConfigs(self, graph: QuantumGraph) -> None:
        """Write configurations for pipeline tasks to butler or check that
        existing configurations are equal to the new ones.

        Parameters
        ----------
        graph : `~lsst.pipe.base.QuantumGraph`
            Execution graph.

        Raises
        ------
        TypeError
            Raised if existing object in butler is different from new data.
        Exception
            Raised if ``extendRun`` is `False` and datasets already exists.
            Content of a butler collection should not be changed if exception
            is raised.
        """

        def logConfigMismatch(msg: str) -> None:
            """Log messages about configuration mismatch."""
            _LOG.fatal("Comparing configuration: %s", msg)

        _LOG.debug("Will save Configs for all tasks")
        # start transaction to rollback any changes on exceptions
        with self.transaction():
            for taskDef in self._task_iter(graph):
                # Config dataset ref is stored in task init outputs, but it
                # may be also be missing.
                task_output_refs = graph.initOutputRefs(taskDef)
                if task_output_refs is None:
                    continue

                config_ref, old_config = self._find_dataset(task_output_refs, taskDef.configDatasetName)
                if config_ref is None:
                    continue

                if old_config is not None:
                    if not taskDef.config.compare(old_config, shortcut=False, output=logConfigMismatch):
                        raise TypeError(
                            f"Config does not match existing task config {taskDef.configDatasetName!r} in "
                            "butler; tasks configurations must be consistent within the same run collection"
                        )
                else:
                    _LOG.debug(
                        "Saving Config for task=%s dataset type=%s", taskDef.label, taskDef.configDatasetName
                    )
                    self.butler.put(taskDef.config, config_ref)

    def savePackageVersions(self, graph: QuantumGraph) -> None:
        """Write versions of software packages to butler.

        Parameters
        ----------
        graph : `~lsst.pipe.base.QuantumGraph`
            Execution graph.

        Raises
        ------
        TypeError
            Raised if existing object in butler is incompatible with new data.
        """
        packages = Packages.fromSystem()
        _LOG.debug("want to save packages: %s", packages)

        # start transaction to rollback any changes on exceptions
        with self.transaction():
            # Packages dataset ref is stored in graph's global init outputs,
            # but it may be also be missing.

            packages_ref, old_packages = self._find_dataset(
                graph.globalInitOutputRefs(), PipelineDatasetTypes.packagesDatasetName
            )
            if packages_ref is None:
                return

            if old_packages is not None:
                # Note that because we can only detect python modules that have
                # been imported, the stored list of products may be more or
                # less complete than what we have now.  What's important is
                # that the products that are in common have the same version.
                _compare_packages(old_packages, packages)
                # Update the old set of packages in case we have more packages
                # that haven't been persisted.
                extra = packages.extra(old_packages)
                if extra:
                    _LOG.debug("extra packages: %s", extra)
                    old_packages.update(packages)
                    # have to remove existing dataset first, butler has no
                    # replace option.
                    self.butler.pruneDatasets([packages_ref], unstore=True, purge=True)
                    self.butler.put(old_packages, packages_ref)
            else:
                self.butler.put(packages, packages_ref)

    def _find_dataset(
        self, refs: Iterable[DatasetRef], dataset_type: str
    ) -> tuple[DatasetRef | None, Any | None]:
        """Find a ref with a given dataset type name in a list of references
        and try to retrieve its data from butler.

        Parameters
        ----------
        refs : `~collections.abc.Iterable` [ `~lsst.daf.butler.DatasetRef` ]
            References to check for matching dataset type.
        dataset_type : `str`
            Name of a dataset type to look for.

        Returns
        -------
        ref : `~lsst.daf.butler.DatasetRef` or `None`
            Dataset reference or `None` if there is no matching dataset type.
        data : `Any`
            An existing object extracted from butler, `None` if ``ref`` is
            `None` or if there is no existing object for that reference.
        """
        ref: DatasetRef | None = None
        for ref in refs:
            if ref.datasetType.name == dataset_type:
                break
        else:
            return None, None

        try:
            data = self.butler.get(ref)
            if data is not None and not self.extendRun:
                # It must not exist unless we are extending run.
                raise ConflictingDefinitionError(f"Dataset {ref} already exists in butler")
        except (LookupError, FileNotFoundError):
            data = None
        return ref, data

    def _task_iter(self, graph: QuantumGraph) -> Iterator[TaskDef]:
        """Iterate over TaskDefs in a graph, return only tasks that have one or
        more associated quanta.
        """
        for taskDef in graph.iterTaskGraph():
            if graph.getNumberOfQuantaForTask(taskDef) > 0:
                yield taskDef

    @contextmanager
    def transaction(self) -> Iterator[None]:
        """Context manager for transaction.

        Default implementation has no transaction support.
        """
        yield


class PreExecInit(PreExecInitBase):
    """Initialization of registry for QuantumGraph execution.

    This class encapsulates all necessary operations that have to be performed
    on butler and registry to prepare them for QuantumGraph execution.

    Parameters
    ----------
    butler : `~lsst.daf.butler.Butler`
        Data butler instance.
    taskFactory : `~lsst.pipe.base.TaskFactory`
        Task factory.
    extendRun : `bool`, optional
        If `True` then do not try to overwrite any datasets that might exist
        in ``butler.run``; instead compare them when appropriate/possible.  If
        `False`, then any existing conflicting dataset will cause a butler
        exception to be raised.
    """

    def __init__(self, butler: Butler, taskFactory: TaskFactory, extendRun: bool = False):
        super().__init__(butler, taskFactory, extendRun)
        self.full_butler = butler
        if self.extendRun and self.full_butler.run is None:
            raise RuntimeError(
                "Cannot perform extendRun logic unless butler is initialized "
                "with a default output RUN collection."
            )

    @contextmanager
    def transaction(self) -> Iterator[None]:
        # dosctring inherited
        with self.full_butler.transaction():
            yield

    def initializeDatasetTypes(self, graph: QuantumGraph, registerDatasetTypes: bool = False) -> None:
        # docstring inherited
        pipeline = graph.taskGraph
        pipelineDatasetTypes = PipelineDatasetTypes.fromPipeline(
            pipeline, registry=self.full_butler.registry, include_configs=True, include_packages=True
        )

        for datasetTypes, is_input in (
            (pipelineDatasetTypes.initIntermediates, True),
            (pipelineDatasetTypes.initOutputs, False),
            (pipelineDatasetTypes.intermediates, True),
            (pipelineDatasetTypes.outputs, False),
        ):
            self._register_output_dataset_types(registerDatasetTypes, datasetTypes, is_input)

    def _register_output_dataset_types(
        self, registerDatasetTypes: bool, datasetTypes: Iterable[DatasetType], is_input: bool
    ) -> None:
        def _check_compatibility(datasetType: DatasetType, expected: DatasetType, is_input: bool) -> bool:
            # These are output dataset types so check for compatibility on put.
            is_compatible = expected.is_compatible_with(datasetType)

            if is_input:
                # This dataset type is also used for input so must be
                # compatible on get as ell.
                is_compatible = is_compatible and datasetType.is_compatible_with(expected)

            if is_compatible:
                _LOG.debug(
                    "The dataset type configurations differ (%s from task != %s from registry) "
                    "but the storage classes are compatible. Can continue.",
                    datasetType,
                    expected,
                )
            return is_compatible

        missing_datasetTypes = set()
        for datasetType in datasetTypes:
            # Only composites are registered, no components, and by this point
            # the composite should already exist.
            if registerDatasetTypes and not datasetType.isComponent():
                _LOG.debug("Registering DatasetType %s with registry", datasetType)
                # this is a no-op if it already exists and is consistent,
                # and it raises if it is inconsistent.
                try:
                    self.full_butler.registry.registerDatasetType(datasetType)
                except ConflictingDefinitionError:
                    if not _check_compatibility(
                        datasetType, self.full_butler.registry.getDatasetType(datasetType.name), is_input
                    ):
                        raise
            else:
                _LOG.debug("Checking DatasetType %s against registry", datasetType)
                try:
                    expected = self.full_butler.registry.getDatasetType(datasetType.name)
                except KeyError:
                    # Likely means that --register-dataset-types is forgotten.
                    missing_datasetTypes.add(datasetType.name)
                    continue
                if expected != datasetType:
                    if not _check_compatibility(datasetType, expected, is_input):
                        raise ValueError(
                            f"DatasetType configuration does not match Registry: {datasetType} != {expected}"
                        )

        if missing_datasetTypes:
            plural = "s" if len(missing_datasetTypes) != 1 else ""
            raise KeyError(
                f"Missing dataset type definition{plural}: {', '.join(missing_datasetTypes)}. "
                "Dataset types have to be registered with either `butler register-dataset-type` or "
                "passing `--register-dataset-types` option to `pipetask run`."
            )


class PreExecInitLimited(PreExecInitBase):
    """Initialization of registry for QuantumGraph execution.

    This class works with LimitedButler and expects that all references in
    QuantumGraph are resolved.

    Parameters
    ----------
    butler : `~lsst.daf.butler.LimitedButler`
        Limited data butler instance.
    taskFactory : `~lsst.pipe.base.TaskFactory`
        Task factory.
    """

    def __init__(self, butler: LimitedButler, taskFactory: TaskFactory):
        super().__init__(butler, taskFactory, False)

    def initializeDatasetTypes(self, graph: QuantumGraph, registerDatasetTypes: bool = False) -> None:
        # docstring inherited
        # With LimitedButler we never create or check dataset types.
        pass
